/*
 * bfs.h
 * LLAMA Graph Analytics
 *
 * Copyright 2014
 *      The President and Fellows of Harvard College.
 *
 * Copyright 2014
 *      Oracle Labs.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#ifndef LL_BFS_H
#define LL_BFS_H

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <float.h>
#include <limits.h>
#include <cmath>
#include <algorithm>
#include <omp.h>

#include "mt_bfs_template.h"
#include "mt_utils.h"
#include "nvgraph.h"


// BFS/DFS definitions for the procedure
//template <class Graph>
template<typename T, typename S>
class bfs_bfs : public ll_bfs_template<T, S> {
public:
    bfs_bfs(Graph<T,S>& _G, T& _root, T& _count, T& _num_thread, T _snp_id)
    : ll_bfs_template<T,S>(_G, _num_thread, _snp_id),
    G(_G), root(_root), count(_count) {}

private:  // list of varaibles
    Graph<T,S>& G;
    T& root;
    T& count;

protected:
    virtual void visit_fw(T v)
    {
        ATOMIC_ADD<T>(&count, 1);
    }

    virtual void visit_rv(T v) {}
    // virtual bool check_navigator(node_t v, edge_t v_idx) {return true;}


};


/**
 * Count vertices in the given BFS traversal
 */
// template <class Graph>
template<typename T, typename S>
class ll_b_bfs {

    T root;
    T num_thread;
    T snp_id;
    Graph<T,S>& G;

public:

    /**
     * Create the benchmark
     *
     * @param r the root
     */
    ll_b_bfs(Graph<T,S>& _G, const T& _r, const T& _num_thread, const T& _snp_id)
		: G(_G), root(_r), num_thread(_num_thread), snp_id(_snp_id) {
    }


    /**
     * Destroy the benchmark
     */
    virtual ~ll_b_bfs(void) {
    }


    /**
     * Run the benchmark
     *
     * @return the numerical result, if applicable
     */
    virtual double run(void) {

        T count = 0;
		    printf("\nroot<%u>, count<%u>, num_thread<%u>, snp_id<%u>\n", root, count, num_thread, snp_id);
        double start = ll_get_time_ms();
        bfs_bfs<T,S> _BFS(G, root, count, num_thread, snp_id);
        _BFS.prepare(root);
        _BFS.do_bfs_forward();
        printf("_bfs_<%f>\n", ll_get_time_ms() - start);

        return count;
    }
};

#endif
