/*
 * NVGraph :
 *
 *
 */

#ifndef BENCHMARK_H
#define BENCHMARK_H

#include "nvgraph.h"
#include "mt_utils.h"

void graph_recovery(int signum){
  uint32_t _num_nodes = 4846609;
  //uint32_t _num_nodes = 41652230;
  vector<float> _pr(_num_nodes, 1.1);
  //unsigned char* infected_bitmap = new unsigned char[(num_nodes+7)/8];
  uint32_t _counter = 5;

  // TIMER START //
  uint32_t tm_start = get_current_time<uint32_t>();
  printf("Failure Signal Detected::%u\n", (uint32_t)signum);

  // Backing up data //
  // (1) reset pointer //
  vector<float> *_pr_ptr = &_pr;
  uint32_t *iter_ptr = &_counter;

  vector<float> new_pr(_num_nodes, 0.0);
  uint32_t resume_iteration = *iter_ptr; (void)resume_iteration;

  // (2) reset nodes status
#pragma omp parallel num_threads(24)
{
  //#pragma omp for schedule(static, 1)
  #pragma omp for nowait
  for(uint32_t nid=1; nid<=_num_nodes; nid++){
    new_pr[nid] = (*_pr_ptr)[nid];
    emulate_latency();
  }
}
  // TIMER STOP //
  uint32_t tm_stop = get_current_time<uint32_t>() - tm_start;
  show_runtime("Failure Recovery Time", tm_stop);

  for(float ft : new_pr) {
    if(ft!=(float)1.1) { printf("check fail\n"); break; }
  }
  // Terminate
  exit(signum);
}

void failure_recovery(int signum){
  uint32_t _num_nodes = 4846609;
  //uint32_t _num_nodes = 41652230;
  vector<float> _pr(_num_nodes, 1.1);
  //unsigned char* infected_bitmap = new unsigned char[(num_nodes+7)/8];
  uint32_t _counter = 5;

  // TIMER START //
  uint32_t tm_start = get_current_time<uint32_t>();
  printf("Failure Signal Detected::%u\n", (uint32_t)signum);

  // Backing up data //
  // (1) reset pointer //
  vector<float> *_pr_ptr = &_pr;
  uint32_t *iter_ptr = &_counter;

  vector<float> new_pr(_num_nodes, 0.0);
  uint32_t resume_iteration = *iter_ptr; (void)resume_iteration;

  // (2) reset nodes status
#pragma omp parallel num_threads(24)
{
  //#pragma omp for schedule(static, 1)
  #pragma omp for nowait
  for(uint32_t nid=1; nid<=_num_nodes; nid++){
    new_pr[nid] = (*_pr_ptr)[nid];
    emulate_latency();
  }
}
  // TIMER STOP //
  uint32_t tm_stop = get_current_time<uint32_t>() - tm_start;
  show_runtime("Failure Recovery Time", tm_stop);

  for(float ft : new_pr) {
    if(ft!=(float)1.1) { printf("check fail\n"); break; }
  }
  // Terminate
  exit(signum);
}

/*
 * node and data (from llama project)
 */
template<typename T, typename S>
struct node_and_data {
	T node; S data;
};

/*
 * compare by data (from llama project)
 */
template<typename T, typename S>
int cmp_by_data_desc(const void* a, const void* b){
	const node_and_data<T,S>* A = (const node_and_data<T,S>*) a;
	const node_and_data<T,S>* B = (const node_and_data<T,S>*) b;

	if(A->data > B->data){ return -1; }
	if(A->data < B->data){ return 1; }
	return (int)(A->node - B->node);
}

/*******************************************************************************
 *
 * PageRank -- Push (from llama project)
 *
 ******************************************************************************/
template<typename T, typename S>
void pr_push(const Graph<T,S>& G,
			 const T iteration,
			 const T snp_id=0,
			 const S threshold=0.000001,
			 const S damping=0.85){

	// //check type
	// T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;
	// (void)MAXX;

  bool single_level = true;
	T num_nodes = 0; // total node(s)
	if(G.snptbl.size()==1){	// single snapshot graph
		num_nodes = G.ntbl.size();
	} else {
		num_nodes = G.snptbl[snp_id].max_id + 1;
    single_level = false;
	} (void) single_level;
	const S N = (S)num_nodes; // total node(s)

	// initialize pagerank value
	vector<S> pr((T)N, (S)0.0);
	vector<S> pr_next((T)N, (S)0.0);

	T counter = 0; // loop counter
	printf("pagerank_num_thread<%d>, iteration<%u>\n", omp_get_max_threads(), iteration);

  do {
		printf("iteration<%u>\n", counter);
    #pragma omp parallel
		{
      #pragma omp for schedule(dynamic,4096)
			for (T t = 1; t < (T)N; t++) {
				T start = 0; T stop = 0;
				get_out_degl(G, t, start, stop);

				T t_degree = stop - start;
				if(t_degree == 0) { continue; }

				S t_pg_rank = pr[t];
				S t_delta = t_pg_rank / t_degree;

				for(T i=start; i<stop; i++){
					if(_ll_get_bit(mem_dram, t)==1) {
						ATOMIC_ADD<S>(&pr_next[G.etbl.dram[i]], t_delta);
					} else {
						ATOMIC_ADD<S>(&pr_next[G.etbl.nvbm[i]], t_delta);
					}
				}
			}

      #pragma omp for schedule(dynamic,4096)
			for (T t = 1; t < (T)N; t++) {
				pr[t] = (1 - damping) / N + damping * pr_next[t];
				pr_next[t] = (S)0.0;
			}
		}

		counter = counter + 1 ;
	} while (counter < iteration);

	/*  print pagerank result
  // convert prop list to node_and_data array
	node_and_data<T, S>* nd = (node_and_data<T, S>*)malloc(pr.size()*sizeof(node_and_data<T, S>));
	nd[0].node = 0; nd[0].data = 0.00000;
	for(T i=1; i<pr.size(); i++){
		nd[i].node = i; nd[i].data = pr[i];
	}

	// sort the result by data in descending order
	qsort(nd, pr.size(), sizeof(node_and_data<T, S>), cmp_by_data_desc<T, S>);

	printf("\tPageRank (Push)\nNID\tPR\n");
	for(T i=0; i<10; i++){ // top 10
		printf("%u\t%f\n", nd[i].node, nd[i].data);
	}
	free(nd);
  */
}

/*******************************************************************************
 *
 * DFS
 *
 ******************************************************************************/
template<typename T, typename S>
void dfs(const Graph<T,S>& G){
	printf("To be updated\n");
}

/*******************************************************************************
 *
 * RP -- Rumor Propagation
 *
 ******************************************************************************/
template<typename T, typename S>
void rp(const Graph<T,S>& G,
        const T snp_id=0){

	//check type
	//T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;
	//(void)MAXX;

  bool single_level = true;
	T num_nodes = 0; // total node(s)
	if(G.snptbl.size()==1){	// single snapshot graph
		num_nodes = G.ntbl.size();
	} else {
		num_nodes = G.snptbl[snp_id].max_id + 1;
    single_level = false;
	} (void) single_level;

	srand(time(NULL)); // seed random number generator
	printf("rumorpropagation_num_thread<%d>\n", omp_get_max_threads());
  T iter = 50;
  for(T i=0; i<iter; i++){
    #pragma omp parallel
    {
      unsigned char* infected_bitmap = new unsigned char[(num_nodes+7)/8];
      #pragma omp for nowait
      for(T i=0; i<((num_nodes+7)/8); i++){
        infected_bitmap[i] = 0;
				if(_ll_get_bit(mem_dram, i)!=1) { emulate_latency(); }
      }

	    queue<T> q; q.push((rand() % (num_nodes - 1)) + 1);
      _ll_set_bit(infected_bitmap, q.front()); // mark visited
			if(_ll_get_bit(mem_dram, q.front())!=1) { emulate_latency(); }

			while(!q.empty()){
				const T src = q.front(); q.pop();

	    	T start = 0; T stop = 0;
				get_out_degl(G, src, start, stop);

				#pragma omp for nowait
				for(T i=start; i<stop; i++){
            T idd = 0; S e_prop = 0.0; if(_ll_get_bit(mem_dram, src)==1) {
						idd = G.etbl.dram[i]; e_prop = G.wtbl.prop_dram[i];
					} else {
						idd = G.etbl.nvbm[i]; e_prop = G.wtbl.prop_nvbm[i];
					}

          if(_ll_get_bit(infected_bitmap, idd)==1) { continue; }
          if((S)rand()/RAND_MAX > e_prop){
						if(_ll_get_bit(mem_dram, idd)!=1) { emulate_latency(); }
            q.push(idd); // add to queue
          }
          _ll_set_bit(infected_bitmap, idd); // mark visited
        } // for
      } // while
    } // #pragma
  } // for
}

/*******************************************************************************
 *
 * randwalk -- Random Walk
 *
 ******************************************************************************/
template<typename T, typename S>
void randwalk(const Graph<T,S>& G,
              const unsigned char* mem_dram,
              const T snp_id=0){

	//check type
	//T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;
	//(void)MAXX;

  bool single_level = true;
	T num_nodes = 0; // total node(s)
	if(G.snptbl.size()==1){	// single snapshot graph
		num_nodes = G.ntbl.size();
	} else {
		num_nodes = G.snptbl[snp_id].max_id + 1;
    single_level = false;
	} (void) single_level;

  T num_iterations = 100000; // number of walks
  const T path_len = 150; // path length
	srand(time(NULL)); // seed random number generator
	printf("randomwalk_num_thread<%d>\n", omp_get_max_threads());
	#pragma omp parallel
  {
    #pragma omp for schedule(static, 1)
    for(T iter=0; iter<num_iterations; iter++){
			uint32_t src = (rand() % (num_nodes - 1)) + 1;
			if(_ll_get_bit(mem_dram, src)==1) { emulate_latency(); }

			if(src==0) { continue; }
      src = get_rand_out_deg(G, src);

			for(uint32_t k=1; k<path_len; k++){
		  	if(src==0) { break; }
        src = get_rand_out_deg(G, src);
				if(_ll_get_bit(mem_dram, src)==1) { emulate_latency(); }
			}
		}
  } // #pragma
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

#endif
