/*
 * NVGraph :
 *
 *
 */

#define INT uint32_t
//#define INT uint64_t
#define FLT float

#include "benchmark.h"
#include "mt_bfs_template.h"
#include "mt_bfs.h"
#include "mt_utils.h"

/*
 * get current time in microseconds
 */
template<typename T>
inline T get_current_time(void){
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return(tv.tv_sec*1000 + tv.tv_usec*0.001);
}

/*
 * convert microseconds to hh:mm:ss:ml
 * // from stackoverflow //
 */
template<typename T>
inline void show_runtime(string mseg, const T timestamp){
  if(!timestamp) { printf("\n%s: 00:00:00:000\n", mseg.c_str()); }

  T milliseconds   = (T) (timestamp) % 1000;
  T seconds    = (((T) (timestamp) - milliseconds)/1000)%60 ;
  T minutes    = (((((T) (timestamp) - milliseconds)/1000) - seconds)/60) %60;
  T hours      = ((((((T) (timestamp) - milliseconds)/1000) - seconds)/60) - minutes)/60;

  printf("\n%s: %02lu:%02lu:%02lu:%03lu\n", mseg.c_str(), (uint64_t)hours, (uint64_t)minutes, (uint64_t)seconds, (uint64_t)milliseconds);
}

/*
template<typename T>
inline void print_utils(int who, struct rusage usage){
  T cpu_times =
  printf("User Time: "); // usage.ru_utime
  printf("System Time: "); // usage.ru_stime
  printf("CPU Time: "); // usage.ru_utime
  printf("CPU Util.: "); // (cputime/#thread)/runtim
}
*/

/*
 * convert timeval to microsecond
 */
template<typename T>
inline T timeval_to_ms(struct timeval t){
  return(t.tv_sec*1000 + t.tv_usec*0.001);
}

/*
 * Estimate input size
 */
template<typename T, typename S>
T calculate_default_sz(const char* fname){
  int fd = open(fname, O_RDONLY, 0);
  if(fd<0){
    perror("open"); abort();
  }
  T f_sz = lseek(fd, 0, SEEK_END);
  T num_xs0 = f_sz/sizeof(xs0<T,S>);
  assert(num_xs0>0);
  return(num_xs0);
}

/*
 * Read the type 1 entry
 */
template<typename T, typename S>
bool get_next(FILE* fd, xs0<T,S>* out){
  ssize_t r = fread(out, sizeof(xs0<T,S>), 1, fd);
  if(r<0){
    perror("fread"); abort();
  }
  return(r>0);
}

/*
 * load nid (stored on dram) use bitwise approach
 *
 */
void load_mem_type(const char* fname, unsigned char* mem_dram){

  if(db==1) { printf("load_mem_type::<%s>\n", fname); }

  ifstream in_file(fname);
  uint32_t nid;
  while(in_file >> nid){
    _ll_set_bit(mem_dram, nid);
  }
  in_file.close();
}

/*
 * load data from file one snapshot per file
 */
template<typename T, typename S>
T load_based_graph(const char* fname,
                   const unsigned char* mem_dram,
                   vector<node<T>>& ntbl,
                   property<S>& wtbl,
                   edge_list<T>& etbl) {

  if(db==1) { printf("load_based_graph::<%s>\n", fname); }

  // // check type
  // T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;

  /*
   * Part 1 : store edge/weight info and count # of node(s) to "etbl"
   */
  FILE* fd = fopen(fname, "rb");
  if(fd==NULL){
    perror("fopen"); abort();
  }

  xs0<T,S> x; T num_nodes = 0; T num_edges = 0;
  T offset = 0; T nvbm_offset = 0; T dram_offset = 0;
  T src = 0; bool mem = NVBM;
  map<T, T> out_deg; pair<T,T> p(0,NVBM);

  while(get_next(fd, &x)){
    if(num_nodes<x.id){ num_nodes = x.id; }
    if(x.weight==(float)-7.3){
      if(_ll_get_bit(mem_dram, x.id)==1){
        mem = DRAM;
      } else {
        mem = NVBM;
      }
      if(src!=0){ // save previous source && # of out_deg
        out_deg.insert(pair<T,T>(src, offset));
      }
      src = x.id;
      offset = nvbm_offset = dram_offset = 0;
    } else {
       num_edges++; offset++;
       if(mem==NVBM){// store on nvbm
        etbl.nvbm.push_back(x.id); emulate_latency();
        wtbl.prop_nvbm.push_back(x.weight); emulate_latency();
      } else { // store on dram
        etbl.dram.push_back(x.id);
        wtbl.prop_dram.push_back(x.weight);
      }
    }
  }
  out_deg.insert(pair<T,T>(src, offset));

  /*
   * Part 2 : update ntbl(snap_id, mem_type, offset)
   */
  offset = 0; nvbm_offset = 0; dram_offset = 0; T len = 0;
  ntbl.push_back(node<T>(0, mem, offset, 0)); // dummy node
  for(T i=1; i<=num_nodes; i++){
    typename map<T,T>::iterator iter = out_deg.find(i);
    if(iter!=out_deg.end()){
      len = iter->second;
      if(_ll_get_bit(mem_dram, iter->first)==1){
        mem = DRAM; offset = dram_offset;
        dram_offset += len;
      } else {
        mem = NVBM; offset = nvbm_offset;
        nvbm_offset += len;
      }
    } else {
      mem = NVBM; offset = nvbm_offset; len = 0;
    }
    ntbl.push_back(node<T>(0, mem, offset, len));
  }
  assert(num_nodes+1==ntbl.size());
  return num_nodes;
}

/*
 * load delta graph info and extend ntbl, etbl, wtbl
 */
template<typename T, typename S>
bool load_delta_graph(const char* fname,
                      vector<snapshot<T,S>>& snptbl){

  if(db==1) { printf("load_delta_graph::add<%s>\n", fname); }

  // check type
  T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;

  const T max_id = snptbl[snptbl.size()-1].max_id; // current max_id

  FILE* fd = fopen(fname, "rb");
  if(fd==NULL){
    perror("fopen"); abort();
  }

  snapshot<T,S> snp;
  //dummy slot for offset to new node(s)
  snp.modified.push_back(_pair<T>(MAXX, MAXX, 0));

  xs0<T,S> x; T num_nodes = 0; T num_edges = 0;
  T offset = 0; T len = 0;
  while(get_next(fd, &x)){
    if(num_nodes<x.id){ num_nodes = x.id; }
    if(x.weight==(float)-7.3){
      // offset points to the first add node(s)
      if(x.id>max_id && snp.modified[0].offset==MAXX){
        snp.modified[0].offset = snp.modified.size();
      }

      if(snp.modified.size()>1){
        snp.modified[snp.modified.size()-1].len = len;
        len = 0;
      }

      snp.modified.push_back(_pair<T>(x.id, offset, 0));
    } else {
      num_edges++; offset++; len++;
      snp.modified_neighbor.push_back(_prop<T,S>(x.id, x.weight));
    }
  }
  snp.modified[snp.modified.size()-1].len = len;
  snp.max_id = num_nodes;
  snptbl.push_back(snp);
  return !snp.modified.empty();
}

/*
 * load data from file one snapshot per file
 */
template<typename T, typename S>
bool extend_based_graph(const unsigned char*  mem_dram,
                        const vector<snapshot<T,S>>& snptbl,
                        vector<node<T>>& ntbl,
                        property<S>& wtbl,
                        edge_list<T>& etbl){

  if(db==1) { printf("extend_based_graph()\n"); }

  // check type
  T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;

  const T snp_id = snptbl.size()-1;
  const T upper_bound_id = snptbl[snp_id].max_id;
  const T lower_bound_id = snptbl[snp_id-1].max_id+1;

  const snapshot<T,S> snp = snptbl[snp_id];
  T cur = snp.modified[0].offset;
  if(cur==MAXX) {  return false; }

  // offset
  T nvbm_offset = etbl.nvbm.size();
  T dram_offset = etbl.dram.size();
  T offset = 0; T len = 0; bool mem = NVBM;
  for(T id=lower_bound_id; id<=upper_bound_id; id++){
    if(cur<snp.modified.size() && id==snp.modified[cur].id){
      if(_ll_get_bit(mem_dram, id)==1){
        mem = DRAM; offset = dram_offset;
      } else {
        mem = NVBM; offset = nvbm_offset;
      }

      // get start and stop offset
      const T start = snp.modified[cur].offset;
      len = snp.modified[cur].len;
      const T stop = start + len;

      for(T i=start; i<stop; i++){
        if(mem==NVBM){
          etbl.nvbm.push_back(snp.modified_neighbor[i].id); emulate_latency();
          wtbl.prop_nvbm.push_back(snp.modified_neighbor[i].prop); emulate_latency();
          nvbm_offset++;
        } else {
          etbl.dram.push_back(snp.modified_neighbor[i].id);
          wtbl.prop_dram.push_back(snp.modified_neighbor[i].prop);
          dram_offset++;
        }
      }
      cur++;
    } else {
      mem = NVBM; offset = nvbm_offset; len = 0;
    }
    ntbl.push_back(node<T>(snp_id, mem, offset, len));
  }
  return true;
}

/*
 * check if node n exist @ snap_id
 */
template<typename T, typename S>
bool inline is_exist(const vector<snapshot<T,S>>& snptbl,
                     const short snp_id,
                     const T n){
  return(snptbl[snp_id].max_id >= n);
}

/*
 * print Graph
 */
template<typename T, typename S>
void print_graph(const Graph_F<T,S>& g){

  // check type
  T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;

  printf("Src\tDst\tProperty\n");
  for(T i=1; i<g.ntbl.size(); i++){
    T hi = MAXX; T low = MAXX; get_ntbl_offset(g.ntbl, g.etbl, i, low, hi);
    for(T l=low; l<hi; l++){
      printf("%u\t", i);
      if(g.ntbl[i].mem_type==DRAM){ // DRAM
        printf("%u\t%f\n", g.etbl.dram[l], g.wtbl.prop_dram[l]);
      } else { // NVBM
        printf("%u\t%f\n", g.etbl.nvbm[l], g.wtbl.prop_nvbm[l]);
      }
    }
  }
}

/*
 * print Graph
 */
template<typename T, typename S>
void print_graph(const Graph_B<T,S>& g){
  printf("print_graph NOT IMPLEMENTED.\n");
}

/*
 * helper function to print edge table (DRAM || NVBM)
 */
template<typename T, typename S>
void print_edge_table_helper(const vector<T>& etbl,
                       const vector<S>& wtbl){
  printf("\tTail\tWeight\n");
  for (T i=0; i<etbl.size(); i++){
    printf("\t%llu\t%f\n", etbl[i], wtbl[i]);
  }
}

/*
 * print edge table ()
 */
template<typename T, typename S>
void print_edge_table(const edge_list<T>& etbl,
                      const property<S>& wtbl){
  printf("DRAM:\n");
  print_edge_table_helper(etbl.dram, wtbl.prop_dram);
  printf("NVBM:\n");
  print_edge_table_helper(etbl.nvbm, wtbl.prop_nvbm);
}

/*
 * print node table
 */
template<typename T>
void print_node_table(const vector<node<T>>& ntbl){
  printf("\nNID\tS.Shot\tMemory\tOffset\n");
  for(T i=1; i<ntbl.size(); i++){
    printf("%llu<%llu>\t%d\t", i, ntbl[i].len, ntbl[i].snapshot);
    if(ntbl[i].mem_type){
      printf("NVBM\t");
    } else {
      printf("DRAM\t");
    }
    printf("%llu\n", ntbl[i].offset);
  } printf("\n");
}

/*
 * print snapshot
 */
template<typename T, typename S>
void print_snapshot(const vector<snapshot<T,S>>& snptbl){
  // // check type
  // T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;

  for(T i=1; i<snptbl.size(); i++){
    printf("\nSnapshot: <%llu>\n\tUpdate Node(s):\n", i);
    for(T j=1; j<snptbl[i].modified.size(); j++){
      // printf("new_id_offset<%llu>\n", snptbl[i].modified[0].offset);
      if(snptbl[i].modified[0].offset==j){
        printf("\tNew Node(s):\n");
      }
      // printf("<<%llu, %llu>>\n", snptbl[i].modified[j].id, snptbl[i].modified[j].offset);
      printf("\t%llu<%llu>:\t", snptbl[i].modified[j].id, snptbl[i].modified[j].len);
      const T start = snptbl[i].modified[j].offset;
      const T stop = j+1==snptbl[i].modified.size() ?
                      snptbl[i].modified_neighbor.size() :  snptbl[i].modified[j+1].offset;
      for(T k=start; k<stop; k++){
        printf("<%llu  %f> ", snptbl[i].modified_neighbor[k].id, snptbl[i].modified_neighbor[k].prop);
      } printf("\n");
    }
  }
}

/*
 * get max_id
 */
template<typename T, typename S>
T get_max_id(const Graph_F<T,S>& g){
  return g.ntbl.size();
}

/*
 * get number of out deg
 */
template<typename T, typename S>
T get_num_out_deg(const Graph<T,S>& g, const T n){
  return (T)g.ntbl[n].len();
}

/*
 * get number of in deg
 */
template<typename T, typename S>
T get_in_degree(const Graph<T,S>& g, const T n){
  printf("get_in_degree NOT IMPLEMENTED.\n");
  return 0;
}

/*
 * get random neighbor of a given node n
 */
template<typename T, typename S>
T get_rand_out_deg(const Graph<T,S>& g, const T n){
  if(g.ntbl[n].len){
    T index = g.ntbl[n].offset + rand()% g.ntbl[n].len;
    if(g.ntbl[n].mem_type==NVBM){
      return g.etbl.nvbm[index];
    } else {
      return g.etbl.dram[index];
    }
  }
  return 0; // nid has no neighbor
}

/*
 * get high/low index of out deg given node n
 * WARNING only work with single snapshot graph
 */
template<typename T, typename S>
void get_out_degl(const Graph<T,S>& g, const T n, T& beg, T& end){
  beg = g.ntbl[n].offset;
  end = beg + g.ntbl[n].len;
  // done
}

/*
 * given snapshot id combine all the snapshot into one
 * NEED to optimize this to run in parallel
 */
template<typename T, typename S>
void flatten(const Graph<T,S>& G, Graph<T,S>& g, const T snp_id){
  if(snp_id>=G.snptbl.size()) {
    printf("Snapshot does not exist...\n");
    return;
  }
  g.id = (short)snp_id;
  const T max_id = G.snptbl[snp_id].max_id;
  T offset = 0; T nvbm_offset = 0; T dram_offset = 0; bool mem = NVBM;
  vector<_prop<T,S>> out_degl;
  for(T i=1; i<=max_id; i++){
    /* not SO SMART here */
    out_degl.clear(); get_out_degl(G, i, snp_id, out_degl);
    for(auto p : out_degl){ // copy etbl, wtbl
      if(G.ntbl[i].mem_type==NVBM){
        g.etbl.nvbm.push_back(p.id); emulate_latency();
        g.wtbl.prop_nvbm.push_back(p.prop); emulate_latency();
      } else {
        g.etbl.dram.push_back(p.id);
        g.wtbl.prop_dram.push_back(p.prop);
      }
    }

    const T len = out_degl.size();
    if(G.ntbl[i].mem_type==NVBM){
      mem = NVBM; offset = nvbm_offset;
      nvbm_offset += len;
    } else {
      mem = DRAM; offset = dram_offset;
      dram_offset += len;
    }
    g.ntbl.push_back(node<T>((short)snp_id, mem, offset, len)); // copy ntbl
  }
}

/*
 * get list of files from given directory
 */
void get_filelist(char* directory, vector<string>& files){

  if(db==1) { printf("get_filelist::directory<%s>\n", directory); }

  struct dirent *ent;
  DIR *dir = opendir(directory);
  if(dir){
    while((ent=readdir(dir))!=NULL){
      if(ent->d_name[0]!='.'){ // skip current and parent dir
        files.push_back(string(directory) + string(ent->d_name));
      }
    }
    closedir(dir);
  } else {
    printf("Cannot open dir <%s>\n", directory);
    exit(1);
  }
  sort(files.begin(), files.end());
}

/*
 * Graph persists.
 */
template<typename T, typename S>
void graph_persist(const Graph_F<T,S>& g){

  // check type
  T MAXX = sizeof(T)==sizeof(uint32_t) ? UINT32_MAX : UINT64_MAX;

  for(T i=1; i<g.ntbl.size(); i++){
    if(!g.ntbl[i].mem_type){
      g.ntbl[i].mem_type=NVBM;
      emulate_latency();
    }
  }

  T dram_len = g.etbl.dram.size();
  T nvbm_len = g.etbl.nvbm.size();
  g.etbl.nvbm.resize(dram_len + nvbm_len);

  for(T i=1; i<g.ntbl.size(); i++){
    T hi = MAXX; T low = MAXX; get_ntbl_offset(g.ntbl, g.etbl, i, low, hi);
    for(T l=low; l<hi; l++){
      if(g.ntbl[i].mem_type==DRAM){
	g.etbl.nvbm.push_back(g.etbl.dram[l]);
      	emulate_latency();
	g.wtbl.nvbm.push_back(g.etbl.dram[l]);
      	emulate_latency();
      } 
    }
  }
}

/*
 * Main function
 */

int main(int argc, char** argv) {
  system("clear");
  time_t tmm = time(NULL); printf("\n%s", ctime(&tmm));

  if(argc<3) {
    printf("\nWARNING: Too few arguments.\n $ ./nvgraph <input_directory> <num_thread> <benchmark-flag>\n");
    printf("\t# benchmark-flag #\n");
    printf("\t0 : run all 4 benchmarks\n");
    printf("\t1 : PageRank\n\t2 : BFS\n\t3 : RandomWalk\n\t4 : RumorPropagation\n\t5 : Flatten\n\n");
    return(1);
  }

  printf("\n\n### Start! ###\n\n");

  // input list
  vector<string> files; get_filelist(argv[1], files);

  INT num_thread = 1; // set thread number
  num_thread = atoi(argv[2]);
  omp_set_dynamic(0);
  omp_set_num_threads(num_thread);

  INT benchmark_flg = atoi(argv[3]);

  // node(s) stored on dram
  const INT total_nodes = 41652230; // twitter_rv.net
  // const INT total_nodes = 4846609; // soc-LiveJournal.net
  mem_dram = new unsigned char[(total_nodes+7)/8];
  #pragma omp parallel num_threads(omp_get_max_threads())
  {
    #pragma omp for nowait
    for(INT i=0; i<((total_nodes+7)/8); i++){
      _ll_clear_bit(mem_dram, i);
    }
  }
  load_mem_type(files[0].c_str(), mem_dram);

  // estimate input size
  INT reserved_sz = calculate_default_sz<INT,FLT>(files[1].c_str());
  // node table
  vector<node<INT>> ntbl; ntbl.reserve(reserved_sz*0.2);
  // edge table (dram, nvbm)
  vector<INT> dr; dr.reserve(reserved_sz*0.1);
  vector<INT> nv; nv.reserve(reserved_sz*0.3);
  edge_list<INT> etbl(dr, nv);
  vector<FLT> p_dr; p_dr.reserve(reserved_sz*0.1);
  vector<FLT> p_nv; p_nv.reserve(reserved_sz*0.3);
  property<FLT> wtbl(p_dr, p_nv);

  struct timeval _tv; gettimeofday(&_tv, NULL);
  string uniq = to_string(_tv.tv_sec);
  string fnmm = uniq + "_soc-LiveJournal1_";
  string fnm_rt_load = fnmm + "load.dat"; // runtime file
  string fnm_rs = fnmm + "resource.dat";  // resource usage file
  string fnm_rt_flt = fnmm + "flatten.dat"; // flatten runtime file
  string fnm_rt_pr = fnmm + "pagerank.dat"; // flatten runtime file
  string fnm_rt_bfs = fnmm + "bfs.dat"; // flatten runtime file
  string fnm_rt_rp= fnmm + "rp.dat"; // flatten runtime file

  vector<uint64_t> wall_clock;
  vector<uint64_t> mem_usage;
  vector<uint64_t> flt_rt;
  vector<uint64_t> pr_rt;
  vector<uint64_t> bfs_rt;
  vector<uint64_t> rp_rt;
  struct rusage r_usage;

  uint64_t start = get_current_time<uint64_t>();
  wall_clock.push_back(get_current_time<uint64_t>());
  getrusage(RUSAGE_SELF, &r_usage);
  mem_usage.push_back(r_usage.ru_maxrss);
  const INT max_id = load_based_graph(files[1].c_str(), mem_dram, ntbl, wtbl, etbl);
  wall_clock.push_back(get_current_time<uint64_t>());
  getrusage(RUSAGE_SELF, &r_usage);
  mem_usage.push_back(r_usage.ru_maxrss);
  uint64_t load_tm = get_current_time<uint64_t>() - start;
  printf("Load Base Graph::Max_ID<%u>", max_id);
  show_runtime("Load Base Graph", load_tm);

  vector<snapshot<INT,FLT>> snptbl;
  snptbl.push_back(snapshot<INT,FLT>(max_id));

  start = get_current_time<uint64_t>();
  if(files.size()>2) {
    for(INT i=2; i<files.size(); i++){
      load_delta_graph(files[i].c_str(), snptbl);
      extend_based_graph(mem_dram, snptbl, ntbl, wtbl, etbl);
      if(i==2 || i==6 || i==11 || i==51 || i==101 || i==201){
        wall_clock.push_back(get_current_time<uint64_t>());
        getrusage(RUSAGE_SELF, &r_usage);
        mem_usage.push_back(r_usage.ru_maxrss);
      }
    }
    uint64_t tmp = get_current_time<uint64_t>() - start;
    load_tm += tmp;
    printf("\nLoad Delta Graph::Max_ID<%u>", snptbl[snptbl.size()-1].max_id);
    show_runtime("Load Delta Graph", tmp);
  }

  show_runtime("Total Load Time", load_tm);

  // Graph structure //
  Graph<INT,FLT> G(ntbl, etbl, wtbl, snptbl);
  const INT num_snapshot = snptbl.size() - 1;
  const INT num_nodes = G.ntbl.size();
  const INT num_edges = G.etbl.dram.size() + G.etbl.nvbm.size();
  printf("\n# Nodes    : %u", num_nodes);
  printf("\n# Edges    : %u", num_edges);
  printf("\n# Snapshot : %lu", G.snptbl.size());
  printf("\n# Threads  : %u", num_thread);

  // number of thread to use //
  vector<INT> threads{1, 2, 4, 8, 16, 24};

  if (benchmark_flg==1 || benchmark_flg==0){
    printf("\n\n### Page Ranking ###\n");
    // Usage: pr_push<T,S>(Graph<T,S>& G, T iteration, T snp_id=0, T threshold=0.000001, S damping=0.85) //
    for (INT thrd : threads){
      if (thrd!=24) { continue; } // only run with 24 threads //
      printf("\nThreads Use: %u.", thrd);
      omp_set_num_threads(thrd);
      start = get_current_time<uint64_t>();
      pr_push<INT,FLT>(G, 10, num_snapshot);
      show_runtime("PageRank", get_current_time<uint64_t>() - start);
    }
  }

  if (benchmark_flg==2 || benchmark_flg==0){
    printf("\n\n### BFS ###\n");
    // Usage: ll_b_bfs<T,S> BFS(Graph<T,S>& G, T root, T num_thread, T snp_id) //
    srand(time(NULL));
    for (INT thrd : threads){
      if (thrd!=24) { continue; } // only run with 24 threads //
      printf("\nThreads Use: %u.", thrd);
      omp_set_num_threads(thrd);
      start = get_current_time<uint64_t>();
	    INT root = (rand() % (num_nodes - 1)) + 1;
      if (argc==5) { root = atoi(argv[4]); } // optional for given root node via commandline //
      ll_b_bfs<INT,FLT> BFS(G, root, num_thread, num_snapshot);
      INT visited_node = (INT)BFS.run();
      printf("Root<%u>\n# Visited nodes<%u>\n", root, visited_node);
      show_runtime("BFS", get_current_time<uint64_t>() - start);
    }
  }

  if (benchmark_flg==3 || benchmark_flg==0){
    printf("\n\n### Random Walk ###\n");
    // Usage: randomwalk<T,S>(Graph<T,S>& G, unsigned char* mem_dram, T snp_id=0) //
    for (INT thrd : threads){
      if (thrd!=24) { continue; } // only run with 24 threads //
      printf("\nThreads Use: %u.", thrd);
      omp_set_num_threads(thrd);
      start = get_current_time<uint64_t>();
      randwalk(G, mem_dram, num_snapshot);
      show_runtime("RandomWalk", get_current_time<uint64_t>() - start);
    }
  }

  if (benchmark_flg==4 || benchmark_flg==0){
    printf("\n\n### Rumor Propagation ###\n");
    // Usage: rp<T,S>(Graph<T,S>& G, T snp_id=0) //
    for (INT thrd : threads){
      if (thrd!=24) { continue; } // only run with 24 threads //
      printf("\nThreads Use: %u.", thrd);
      omp_set_num_threads(thrd);
      start = get_current_time<uint64_t>();
      rp<INT,FLT>(G, num_snapshot);
      show_runtime("RumorPropagation", get_current_time<uint64_t>() - start);
    }
  }

  if (benchmark_flg==5){
    printf("\n### FLATTEN NOT IMPLEMENTED ON SINGLE SNAPSHOT ###\n");
  //    start = get_current_time<uint64_t>();
  //   vector<node<INT>> gntbl1; gntbl1.reserve(reserved_sz*0.2);
  //   vector<INT> gdr1; gdr1.reserve(reserved_sz*0.1);
  //   vector<INT> gnv1; gnv1.reserve(reserved_sz*0.3);
  //   edge_list<INT> getbl1(gdr1, gnv1);
  //   vector<FLT> g_dr1; g_dr1.reserve(reserved_sz*0.1);
  //   vector<FLT> g_nv1; g_nv1.reserve(reserved_sz*0.3);
  //   property<FLT> gwtbl1(g_dr1, g_nv1);
  //   Graph<INT,FLT> g1(gntbl1, getbl1, gwtbl1);
  //   flatten_th(G, g1, num_snapshot); //
  //   show_runtime("flatten_multi_th()", get_current_time<uint64_t>() - start);
  }

  printf("\n### Completed! ###\n\n\n");

  return(0);
}
